/* intfftw3 : scilab interface for fftw library
 *
 * contents :
 *
 *  Utilities functions :
 *  -------------------
 *
 *   - int PushFTTWPlan(void)
 *   - int UpdateFTTWPlan(fftw_plan p, int rank, const fftw_iodim *dims,
 *                        int howmany_rank, const fftw_iodim *howmany_dims,
 *                        int isn)
 *   - int FreeFTTWPlan(void)
 *   - fftw_plan GetFFTWPlan(int rank, const fftw_iodim *dims,
 *                           int howmany_rank, const fftw_iodim *howmany_dims,
 *                           double *ri, double *ii, double *ro, double *io,
 *                           unsigned flags)
 *
 *  Interfacing functions :
 *  ---------------------
 *
 *   - int intfftw_c(char *fname)
 *   - int intsetfftwflags_c(char *fname)
 *   - int intsetfftwinplace_c(char *fname)
 *   - int intsetfftwnorm_c(char *fname)
 *   - int intpushfftwplan_c(char *fname)
 *   - int intfreefftwplan_c(char *fname)
 *
 *  alan.layec@inria.fr, initial revision  : 30/09/06
 *  alan, use of GetData() to disable outgoing bug
 *          (because of intersci common)   : 05/10/06
 *  alan, debug fftwpush_plan              : 10/10/06
 *
 * TODO : interface for fftw_set_timelimit
 *
 */

#include "intfftw3.h"
/* Copyright INRIA */

/* definition of a static structure to store parameters
 * of FFTW plan - set here default value -
 */
static FFTW_Plan_struct Sci_Plan =
{
  NULL,          /* fftw_plan *p              */
  NULL,          /* int *rank                 */
  NULL,          /* fftw_iodim **dims         */
  NULL,          /* int *howmany_rank         */
  NULL,          /* fftw_iodim **howmany_dims */
  NULL,          /* int *isn                  */
  NULL,          /* unsigned *sflags          */

  {NULL,NULL,    /* double * input/output     */
     NULL,NULL},
  FFTW_ESTIMATE, /* unsigned flags            */
  1,             /* int in                    */
  1,             /* int norm                  */

  0,             /* nb_CPlan                  */
};

/**********************
 * Utilities functions
 **********************/

/* Put fftw_plan p value in the static structure Sci_Plan
 * and parameters of fftw_plan_guru_split_dft to reuse
 * plan if parameters are the same between each call of
 * fftw().
 *
 * Input : void
 *
 * Output : int, return 0 if failed, else 1.
 *
 */
int PushFTTWPlan(void)
{
  /* declaration of local variable */
  int nbp,i;

  /* get number of plan */
  nbp = Sci_Plan.nb_CPlan+1;

  /* realloc Sci_Plan.p */
  if ((Sci_Plan.p=(fftw_plan *) REALLOC(Sci_Plan.p,sizeof(fftw_plan)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.p[nbp]=Sci_Plan.p[nbp-1];

  /* realloc Sci_Plan.rank */
  if ((Sci_Plan.rank=(int *) REALLOC(Sci_Plan.rank,sizeof(int)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.rank[nbp]=Sci_Plan.rank[nbp-1];

  /* realloc Sci_Plan.dims */
  if ((Sci_Plan.dims=(fftw_iodim **) REALLOC(Sci_Plan.dims,sizeof(fftw_iodim *)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.dims[nbp] = NULL;
  if (Sci_Plan.dims[nbp-1]!=NULL)
  {
   if ((Sci_Plan.dims[nbp]=(fftw_iodim *) MALLOC(sizeof(fftw_iodim)*Sci_Plan.rank[nbp])) == NULL)
      return(0);
   for(i=0;i<Sci_Plan.rank[nbp];i++)
   {
     Sci_Plan.dims[nbp][i].n  = Sci_Plan.dims[nbp-1][i].n;
     Sci_Plan.dims[nbp][i].is = Sci_Plan.dims[nbp-1][i].is;
     Sci_Plan.dims[nbp][i].os = Sci_Plan.dims[nbp-1][i].os;
   }
  }

  /* realloc Sci_Plan.howmany_rank */
  if ((Sci_Plan.howmany_rank=(int *) REALLOC(Sci_Plan.howmany_rank,sizeof(int)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.howmany_rank[nbp]=Sci_Plan.howmany_rank[nbp-1];

  /* realloc Sci_Plan.howmany_dims */
  if ((Sci_Plan.howmany_dims=(fftw_iodim **) REALLOC(Sci_Plan.howmany_dims,sizeof(fftw_iodim *)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.howmany_dims[nbp]=NULL;
  if (Sci_Plan.howmany_dims[nbp-1]!=NULL)
  {
   if ((Sci_Plan.howmany_dims[nbp]=(fftw_iodim *) MALLOC(sizeof(fftw_iodim)*Sci_Plan.howmany_rank[nbp])) == NULL)
      return(0);
   for(i=0;i<Sci_Plan.howmany_rank[nbp];i++)
   {
     Sci_Plan.howmany_dims[nbp][i].n  = Sci_Plan.howmany_dims[nbp-1][i].n;
     Sci_Plan.howmany_dims[nbp][i].is = Sci_Plan.howmany_dims[nbp-1][i].is;
     Sci_Plan.howmany_dims[nbp][i].os = Sci_Plan.howmany_dims[nbp-1][i].os;
    }
  }

  /* realloc Sci_Plan.sflags */
  if ((Sci_Plan.sflags=(unsigned *) REALLOC(Sci_Plan.sflags,sizeof(unsigned)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.sflags[nbp]=Sci_Plan.sflags[nbp-1];

  /* realloc Sci_Plan.isn */
  if ((Sci_Plan.isn=(int *) REALLOC(Sci_Plan.isn,sizeof(int)*(nbp+1))) == NULL)
     return(0);
  Sci_Plan.isn[nbp]=Sci_Plan.isn[nbp-1];

  /* increase number of plan */
  Sci_Plan.nb_CPlan++;

  /* */
  return(1);
}

/* Free plan the Sci_Plan structure
 *
 * Input : void
 *
 * Output : int, return 0 if failed, else 1.
 *
 */
int FreeFTTWPlan(void)
{
  /* declaration of local variable */
  int i,nbp;

  /* get number of plan */
  nbp = Sci_Plan.nb_CPlan+1;

  if (Sci_Plan.p!=NULL)
  {
   /* free Sci_Plan.p */
   FREE(Sci_Plan.p);
   Sci_Plan.p=NULL;

   /* free Sci_Plan.dims/Sci_Plan.howmany_dims */
   for (i=0;i<nbp;i++)
   {
    if (Sci_Plan.dims[i]!=NULL)
    {
     FREE(Sci_Plan.dims[i]);
    }
    if (Sci_Plan.howmany_dims[i]!=NULL)
    {
     FREE(Sci_Plan.howmany_dims[i]);
    }
   }
   FREE(Sci_Plan.dims);
   FREE(Sci_Plan.howmany_dims);
   Sci_Plan.dims=NULL;
   Sci_Plan.howmany_dims=NULL;

   /* free Sci_Plan.sflags */
   FREE(Sci_Plan.sflags);
   Sci_Plan.sflags=NULL;

   /* free Sci_Plan.rank */
   FREE(Sci_Plan.rank);
   Sci_Plan.rank=NULL;

   /* free Sci_Plan.howmany_rank */
   FREE(Sci_Plan.howmany_rank);
   Sci_Plan.howmany_rank=NULL;

   /* free Sci_Plan.isn */
   FREE(Sci_Plan.isn);
   Sci_Plan.isn=NULL;

   /* number of plan = 0 */
   Sci_Plan.nb_CPlan=0;
  }
  /* */
  return(1);
}

/* Update the static structure Sci_Plan with new
 * fftw_plan p value and parameters of
 * fftw_plan_guru_split_dft to reuse
 * plan if parameters are the same between each call of
 * fftw().
 *
 * Input : fftw_plan p
 *         int rank
 *         const fftw_iodim *dims
 *         int howmany_rank
 *         const fftw_iodim *howmany_dims
 *
 * Output : int, return 0 if failed, else 1.
 *
 */
int UpdateFTTWPlan(fftw_plan p, int rank, const fftw_iodim *dims,
                   int howmany_rank, const fftw_iodim *howmany_dims, int isn)
{
  /* declaration of local variable */
  int curp,i;

  /* get current plan number*/
  curp = Sci_Plan.nb_CPlan;

  /* Sci_Plan.p */
  if (Sci_Plan.p==NULL)
  {
   if ((Sci_Plan.p=(fftw_plan *) MALLOC(sizeof(fftw_plan))) == NULL)
       return(0);
  }
  Sci_Plan.p[curp]=p;

  /* Sci_Plan.rank */
  if (Sci_Plan.rank==NULL)
  {
   if ((Sci_Plan.rank=(int *) MALLOC(sizeof(int))) == NULL)
      return(0);
  }
  /* Sci_Plan.dims */
  if (Sci_Plan.dims==NULL)
  {
   if ((Sci_Plan.dims=(fftw_iodim **) MALLOC(sizeof(fftw_iodim *))) == NULL)
     return(0);
   Sci_Plan.dims[0]=NULL;
  }
  if (Sci_Plan.dims[curp]==NULL)
  {
    if ((Sci_Plan.dims[curp]=(fftw_iodim *) MALLOC(sizeof(fftw_iodim)*rank)) == NULL)
      return(0);
  }
  else
  {
   if (Sci_Plan.rank[curp] < rank)
   {
    if ((Sci_Plan.dims[curp]=(fftw_iodim *) REALLOC(Sci_Plan.dims[curp],sizeof(fftw_iodim)*rank)) == NULL)
       return(0);
   }
  }
  Sci_Plan.rank[curp]=rank;
  for(i=0;i<rank;i++)
  {
   if (dims!=NULL)
   {
    Sci_Plan.dims[curp][i].n  = dims[i].n;
    Sci_Plan.dims[curp][i].is = dims[i].is;
    Sci_Plan.dims[curp][i].os = dims[i].os;
   }
  }

  /* Sci_Plan.howmany_rank */
  if (Sci_Plan.howmany_rank==NULL)
  {
    if ((Sci_Plan.howmany_rank=(int *) MALLOC(sizeof(int))) == NULL)
      return(0);
  }
  /* Sci_Plan.howmany_dims */
  if (Sci_Plan.howmany_dims==NULL)
  {
   if ((Sci_Plan.howmany_dims=(fftw_iodim **) MALLOC(sizeof(fftw_iodim *))) == NULL)
      return(0);
   Sci_Plan.howmany_dims[0]=NULL;
  }
  if (Sci_Plan.howmany_dims[curp]==NULL)
  {
    if (howmany_rank!=0)
    {
     if ((Sci_Plan.howmany_dims[curp]=(fftw_iodim *) MALLOC(sizeof(fftw_iodim)*howmany_rank)) == NULL)
       return(0);
    }
  }
  else
  {
   if (Sci_Plan.howmany_rank[curp] < howmany_rank)
   {
    if ((Sci_Plan.howmany_dims[curp]=(fftw_iodim *) REALLOC(Sci_Plan.howmany_dims[curp],sizeof(fftw_iodim)*howmany_rank)) == NULL)
       return(0);
   }
  }
  Sci_Plan.howmany_rank[curp]=howmany_rank;
  for(i=0;i<howmany_rank;i++)
  {
   if (howmany_dims!=NULL)
   {
    Sci_Plan.howmany_dims[curp][i].n  = howmany_dims[i].n;
    Sci_Plan.howmany_dims[curp][i].is = howmany_dims[i].is;
    Sci_Plan.howmany_dims[curp][i].os = howmany_dims[i].os;
   }
  }

  /* Sci_Plan.sflags */
  if (Sci_Plan.sflags==NULL)
  {
   if ((Sci_Plan.sflags=(unsigned *) MALLOC(sizeof(unsigned))) == NULL)
      return(0);
  }
  Sci_Plan.sflags[curp]=Sci_Plan.flags;

  /* Sci_Plan.isn */
  if (Sci_Plan.isn==NULL)
  {
     if ((Sci_Plan.isn=(int *) MALLOC(sizeof(int))) == NULL)
     return(0);
  }
  Sci_Plan.isn[curp]=isn;

  /* */
  return(1);
}


/* Return a valid plan ptr.
 * This function search in the static structure if
 * the given input parameters follows an already stored
 * set of parameters.
 * If found then return an already stored plan ptr.
 * If not found, then clean the last index and create
 * a new set of parameters (and a new plan)
 * with fftw_plan_guru_split_dft.
 *
 * Input : int rank
 *         const fftw_iodim *dims
 *         int howmany_rank
 *         const fftw_iodim *howmany_dims
 *         double *ri, double *ii
 *         double *ro, double *io
 *         unsigned flags, int isn
 *
 * Output : fftw_plan
 *
 *
 */
fftw_plan GetFFTWPlan(int rank, const fftw_iodim *dims,
                      int howmany_rank, const fftw_iodim *howmany_dims,
                      double *ri, double *ii, double *ro, double *io,
                      unsigned flags, int isn)
{
  /* declaration of local variable */
  fftw_plan p;
  int i,j,ok,ko;

  ok=0;
  ko=ok;

  /* check if for set of parameters there is plan stored
   * in Sci_Plan
   */
  if (Sci_Plan.p!=NULL)
  {
   for (i=0;i<(Sci_Plan.nb_CPlan+1);i++)
   {/* set ok flag to false.
     * ok becomes 1 if it exists a plan for given parameters
     * in Sci_pplan struct.
     */
    ok=0;

    if (i==Sci_Plan.nb_CPlan)
    {
     if (Sci_Plan.sflags[i]!=Sci_Plan.flags) /* ! must be removed later ! */
           break;
    }
    /* check if Sci_Plan.p[i] is already allocated */
    if (Sci_Plan.p[i]!=NULL)
    {
      /* check if Sci_Plan.rank[i] */
      if (Sci_Plan.rank[i]==rank)
      {
       /* check if Sci_Plan.howmany_rank[i] */
       if (Sci_Plan.howmany_rank[i]==howmany_rank)
       {
        /* check if Sci_Plan.isn[i] */
        if (Sci_Plan.isn[i]==isn)
        {
         /* check for NULL case of dims */
         if (dims==Sci_Plan.dims[i]) /*NULL case*/
         {
          /* check for NULL case of howmany_dims */
          if(howmany_dims==Sci_Plan.howmany_dims[i]) /*NULL case*/
              ok=1; /*it is stupid!*/
         }
         else
         {
          /* set ok to true as default */
          ok=1 ;
          /* check dims (of type fftw_iodim) structure */
          for (j=0;j<Sci_Plan.rank[i];j++)
          {
           if (Sci_Plan.dims[i][j].n != dims[j].n)   {ok=0;break;}
           if (Sci_Plan.dims[i][j].is != dims[j].is) {ok=0;break;}
           if (Sci_Plan.dims[i][j].os != dims[j].os) {ok=0;break;}
          }
          /* that ok, all seems to be good */
          if (ok)
          {
           /*NULL case*/
           if(howmany_dims==Sci_Plan.howmany_dims[i]) 
           {
            /* */
            ok=1;
            /* */
            p = Sci_Plan.p[i];
            /* get out from Sci_Plan loop */
            break;
           }
           else
           {
            /* check howmany_dims (of type fftw_iodim) structure */
            for (j=0;j<Sci_Plan.howmany_rank[i];j++)
            {
             if (Sci_Plan.howmany_dims[i][j].n != howmany_dims[j].n)   {ok=0;break;}
             if (Sci_Plan.howmany_dims[i][j].is != howmany_dims[j].is) {ok=0;break;}
             if (Sci_Plan.howmany_dims[i][j].os != howmany_dims[j].os) {ok=0;break;}
            }
            if (ok)
            {
             /* */
             p = Sci_Plan.p[i];
             /* get out from Sci_Plan loop */
             break;
            }
           }
          }
         }
        }
       }
      }
    }
   }
  }

  /* if !ok then destroy the last set of parameters
   * and create a new set at index Sci_Plan.nb_CPlan
   */
  if (!ok)
  {
   if (Sci_Plan.p!=NULL)
   {
    if (Sci_Plan.p[Sci_Plan.nb_CPlan]!=NULL)
    {
      /* */
      fftw_destroy_plan(Sci_Plan.p[Sci_Plan.nb_CPlan]);
    }
   }

   /* create a new plan */
   p = fftw_plan_guru_split_dft(rank, dims,howmany_rank, howmany_dims,
                                ri, ii, ro, io,flags);

   /* */
   Sci_Plan.io[0]=ri;
   Sci_Plan.io[1]=ii;
   Sci_Plan.io[2]=ro;
   Sci_Plan.io[3]=io;

   /* store new plan and its parameters in Sci_Plan */
   if ((UpdateFTTWPlan(p, rank, dims, howmany_rank, howmany_dims, isn))==0)
   {
    /* */
    fftw_destroy_plan(p);
    /* */
    p=NULL;
   }
  }

  /* return value of fftw_plan */
  return (p);
}

/* fftw function.
 *
 * Scilab Calling sequence :
 *   -->fftw(a,sign,dim,incr);
 *
 * Input : rhs(1) : a scilab double complex or real
 *                  vector (row or column)/matrix
 *
 *         rhs(2) : a scilab double or integer
 *                  scalar (-1 or 1) giving the sign
 *                  in the exponential component
 *
 *         rhs(3) : a scilab double or integer
 *                  vector (row or column) of dimension
 *                  of the Fast Fourier Transform to perform
 *
 *         rhs(4) : a scilab double or integer
 *                  scalar...
 *
 * Output : a scilab double complex or real
 *          vector(row or column)/matrix that
 *          gives the result of the transform.
 *
 */
int intfftw_c(char *fname)
{
  /* declaration of variables to store scilab parameters address */
  static int lr1,li1,it1,m1, n1; /* Rhs(1) */

  static int lr2,li2,it2,m2,n2;   /* Rhs(2) */

  static int it3,m3,n3;           /* Rhs(3) */
  int mn3;

  static int it4,m4,n4;           /* Rhs(4) */
  int mn4;

  int *header;

  double         *ptr_d;
  char           *ptr_c;
  unsigned char  *ptr_uc;
  short          *ptr_s;
  unsigned short *ptr_us;
  int            *ptr_i;
  unsigned int   *ptr_ui;

  /* specific declaration for FFTW library variable */
  fftw_plan p;
  int rank;
  fftw_iodim *dims;
  int howmany_rank;
  fftw_iodim *howmany_dims;

  /* input/output address for transform variables */
  double *ri,*ii,*ro,*io;
  double *ptr;

  /* local counter variable */
  int i,j,k;

  /* local variable */
  int isn,vect;
  int *n,*nspn;
  double zero=0;

  /****************************************
   * Basic constraints on rhs parameters  *
   ****************************************/

  /* check min/max lhs/rhs parameters of scilab funcion */
  CheckRhs(1,4);
  CheckLhs(1,1);

  /* 3 rhs not allowed */
  if (Rhs==3)
  {
   Scierror(39,"%s: "
               "Bad number of Rhs parameters\n",
               fname);

   return(0);
  }

  /* retrieve address of input variable to transform */
  GetRhsCVar(1, "d", &it1, &m1, &n1, &lr1, &li1);

  /* void input gives void output */
  if ((m1<1)|(n1<1))
  {
   LhsVar(1)=1;
   return(0);
  }

  /******************************************
   * Advanced constraints on rhs parameters *
   * Get value of rhs parameters            *
   ******************************************/

  /*only one rhs parameter*/
  /*forward fft*/
  if (Rhs==1)
  {
   /* sign of the exp. component */
   isn=-1;
  }
  /* two or four rhs parameters case */
  else
  {
   /* Get dim/type of Rhs(2) */
   header = GetData(2);
   m2  = header[1];
   n2  = header[2];
   it2 = header[3];

   /* */
   CheckDims(2,m2,n2,1,1);

   /* look at for type of Rhs(2) */
   if (VarType(2)==8) /* int */
   {
    /* */
    switch (it2)
    {
     case I_CHAR   : ptr_c=IC_CHAR(&header[4]);
                     isn=(int) ptr_c[0];
                     break;
     case I_INT16  : ptr_s=IC_INT16(&header[4]);
                     isn=(int) ptr_s[0];
                     break;
     case I_INT32  : ptr_i=IC_INT32(&header[4]);
                     isn=(int) ptr_i[0];
                     break;
     case I_UCHAR  : ptr_uc=IC_UCHAR(&header[4]);
                     isn=(int) ptr_uc[0];
                     break;
     case I_UINT16 : ptr_us=IC_UINT16(&header[4]);
                     isn=(int) ptr_us[0];
                     break;
     case I_UINT32 : ptr_ui=IC_UINT32(&header[4]);
                     isn=(int) ptr_ui[0];
                     break;
    }
   }
   else if (VarType(2)==1)  /* double */
   {
    /* */
    ptr_d=(double *)(&header[4]);
    /* */
    isn=(int)ptr_d[0];
   }
   /* if is not int/double then error message */
   else
   {
    Scierror(53,"%s: "
                "Bad type for Rhs(2) parameter\n",
                fname);

    return(0);
   }

   /* check value of second rhs argument */
   if ((isn!=1)&&(isn!=-1))
   {
    Scierror(53,"%s: "
                "Bad value for Rhs(2) parameter\n",
                fname);

    return(0);
   }

   /* four rhs parameters */
   if (Rhs==4)
   {
    /* Get dim/type of Rhs(3) */
    header = GetData(3);
    m3  = header[1];
    n3  = header[2];
    it3 = header[3];

    /* look at for type of Rhs(3) */
    if ((VarType(3)!=8)&&(VarType(3)!=1))
    {
     Scierror(53,"%s: "
                 "Bad type for Rhs(3) parameter\n",
                 fname);

     return(0);
    }

    /* */
    mn3=m3*n3;

    /* check dims */
    if (m3*n3==0)
    {
     Scierror(999,"%s: "
                  "Bad size for Rhs(3) parameter\n",
                  fname);

     return(0);
    }

    /* Get dim/type of Rhs(4) */
    header = GetData(4);
    m4  = header[1];
    n4  = header[2];
    it4 = header[3];

    /* look at for type of Rhs(4) */
    if ((VarType(4)!=8)&&(VarType(4)!=1)) /* int */
    {
     Scierror(53,"%s: "
                 "Bad type for Rhs(4) parameter\n",
                 fname);

     return(0);
    }

    /* */
    mn4=m4*n4;

    /* check dims */
    if (m4*n4==0)
    {
     Scierror(999,"%s: "
                  "Bad size for Rhs(4) parameter\n",
                  fname);

     return(0);
    }

    /* cross variable size checking */
    if (mn4!=mn3)
    {
     Scierror(999,"%s: "
                  "Size of Rhs(3) and Rhs(4) parameters "
                  "must be the same\n",
                   fname);
     return(0);
    }

    /* alloc n/nspn with MALLOC */
    if ((n=(int *)MALLOC(mn4*sizeof(int)))==NULL)
    {
     /* */
     Scierror(999,"%s: "
                  "Allocation memory error\n",
                  fname);
     return(0);
    }

    if ((nspn=(int *)MALLOC(mn4*sizeof(int)))==NULL)
    {
     /* */
     Scierror(999,"%s: "
                  "Allocation memory error\n",
                  fname);
     FREE(n);

     return(0);
    }

    /* n    <- Rhs(3) */
    /* nspn <- Rhs(4) */
    /* check values   */
    for (i=0;i<mn3;i++)
    {
     /* Rhs(3) */
     header = GetData(3);

     switch (VarType(3))
     {
      /* double */
      case 1 :
      {
       ptr_d=(double *)(&header[4]);
       n[i]=(int)ptr_d[i];
       break;
      }
      /* int */
      case 8 :
      {
       switch (it3)
       {
        case I_CHAR   : ptr_c=IC_CHAR(&header[4]);
                        n[i]=(int) ptr_c[i];
                        break;

        case I_INT16  : ptr_s=IC_INT16(&header[4]);
                        n[i]=(int) ptr_s[i];
                        break;

        case I_INT32  : ptr_i=IC_INT32(&header[4]);
                        n[i]=(int) ptr_i[i];
                        break;

        case I_UCHAR  : ptr_uc=IC_UCHAR(&header[4]);
                        n[i]=(int) ptr_uc[i];
                        break;

        case I_UINT16 : ptr_us=IC_UINT16(&header[4]);
                        n[i]=(int) ptr_us[i];
                        break;

        case I_UINT32 : ptr_ui=IC_UINT32(&header[4]);
                        n[i]=(int) ptr_ui[i];
                        break;
       }
       break;
      }
     }
     /* check value of n[i] */
     if (n[i]<=0)
     {
      Scierror(999,"%s: "
                   "Bad value for Rhs(3) parameter : %d\n",
                    fname,n[i]);
      FREE(n);FREE(nspn);
      return(0);
      break;
     }

     /* Rhs(4) */
     header = GetData(4);

     switch (VarType(4))
     {
      /* double */
      case 1 :
      {
       ptr_d=(double *)(&header[4]);
       nspn[i]=(int)ptr_d[i];
       break;
      }
      /* int */
      case 8 :
      {
       switch (it4)
       {
        case I_CHAR   : ptr_c=IC_CHAR(&header[4]);
                        nspn[i]=(int) ptr_c[i];
                        break;

        case I_INT16  : ptr_s=IC_INT16(&header[4]);
                        nspn[i]=(int) ptr_s[i];
                        break;

        case I_INT32  : ptr_i=IC_INT32(&header[4]);
                        nspn[i]=(int) ptr_i[i];
                        break;

        case I_UCHAR  : ptr_uc=IC_UCHAR(&header[4]);
                        nspn[i]=(int) ptr_uc[i];
                        break;

        case I_UINT16 : ptr_us=IC_UINT16(&header[4]);
                        nspn[i]=(int) ptr_us[i];
                        break;

        case I_UINT32 : ptr_ui=IC_UINT32(&header[4]);
                        nspn[i]=(int) ptr_ui[i];
                        break;
       }
       break;
      }
     }
     /* check value of nspn[i] */
     if (nspn[i]<=0)
     {
      Scierror(999,"%s: "
                   "Bad value for Rhs(4) parameter : %d\n",
                    fname,nspn[i]);
      FREE(n);FREE(nspn);
      return(0);
      break;
     }

    }
   }
  }

  /*********************************************
   * Set address of input/ouput array for fftw *
   *********************************************/

  /* complex data */
  if (it1==1)
  {
   /* set input array address */
   ri=stk(lr1);
   ii=stk(li1);
  }
  /* real data */
  else if (it1==0)
  {
   /* set input array address */
   CreateCVar(1, "d", (i=1,&i), &m1, &n1, &lr1,&li1);

   ri=stk(lr1);
   ii=stk(li1);

   /* set all elements of imaginary parts to 0 */

   C2F(dset)((i=m1*n1,&i), &zero, ii, (k=1,&k));
//    for (i=0;i<m1*n1;i++) ii[i]=0;
  }

  /* set output array address */
  /* in place fft*/
  if (Sci_Plan.in)
  {
   ro=ri;
   io=ii;
  }
  /* out place fft*/
  else
  {
   /* */
   CreateCVar(2, "d", &it1, &m1, &n1, &lr2,&li2);

   /* */
   ro=stk(lr2);
   io=stk(li2);
  }

  /* reverse address of input/output array
   * if it is a backward fft 
   */
  if (isn==1)
  {
   /* reverse input */
   ptr=ri;
   ri=ii;
   ii=ptr;

   /* reverse output */
   ptr=ro;
   ro=io;
   io=ptr;
  }

  /**********************************
   * Set parameters needed for fftw *
   **********************************/

  /* check if it is one column or row vector */
  vect=0;
  if ((m1==1)|(n1==1)) vect=1;

  /* vector transform */
  if ((vect)&&(Rhs<=2))
  {
   /* set parameters of fftw_plan_guru_split_dft */
   rank=1;
   if ((dims=(fftw_iodim *)MALLOC(sizeof(fftw_iodim)))==NULL)
   {
    /* */
    Scierror(999,"%s: "
                 "Allocation memory error\n",
                 fname);
    /* */
    return(0);
   }

   dims[0].n = m1*n1;
   dims[0].is = 1;
   dims[0].os = 1;

   howmany_rank=0;
   howmany_dims=NULL;
  }
  else
  {
   /* 2D fft */
   if (Rhs<=2)
   {
    /* set parameters of fftw_plan_guru_split_dft */
    rank=2;
    if ((dims=(fftw_iodim *)MALLOC(sizeof(fftw_iodim)*rank))==NULL)
    {
     /* */
     Scierror(999,"%s: "
                  "Allocation memory error\n",
                  fname);
     /* */
     return(0);
    }

    dims[0].n = m1;
    dims[0].is = 1;
    dims[0].os = 1;

    dims[1].n = n1;
    dims[1].is = m1;
    dims[1].os = m1;

    howmany_rank=0;
    howmany_dims=NULL;
   }
   /* multidimensional fft */
   else if (Rhs==4)
   {
    /* size 1x1 for Rhs(3)/Rhs(4) */
    if (mn3==1)
    /* **This is a special case**
     * all is made in the following block.
     * compatibility with scilab fft function.
     */
    {
     /* set parameters of fftw_plan_guru_split_dft */
     rank=1;
     if ((dims=(fftw_iodim *)MALLOC(sizeof(fftw_iodim)*rank))==NULL)
     {
      /* */
      Scierror(999,"%s: "
                   "Allocation memory error\n",
                   fname);
      /* */
      return(0);
     }

     dims[0].n = n[0];
     dims[0].is = nspn[0];
     dims[0].os = nspn[0];

     howmany_rank=1;
     if ((howmany_dims=(fftw_iodim *)MALLOC(sizeof(fftw_iodim)*howmany_rank))==NULL)
     {
      /* */
      Scierror(999,"%s: "
                   "Allocation memory error\n",
                   fname);
      /* */
      return(0);
     }

     /* find number of transforms to compute */
     howmany_dims[0].n = 0;
     j = (n[0]-1)*nspn[0];
     while(j<((m1*n1)-(nspn[0]-1)))
     {
      howmany_dims[0].n++;
      j += nspn[0]*n[0];
     }
     if (j<=((m1*n1)-(nspn[0]-1)))
       k=j+nspn[0];
     else
       k=j-nspn[0]*n[0]+nspn[0];

     howmany_dims[0].is = n[0]*nspn[0];
     howmany_dims[0].os = n[0]*nspn[0];

     if ((p = GetFFTWPlan(rank, dims,
                          howmany_rank, howmany_dims,
                          ri, ii, ro, io,
                          Sci_Plan.flags,isn)) == NULL)
     {
      /* */
      Scierror(999,"%s: "
                   "Allocation memory error\n",
                   fname);
      /* */
      FREE(dims);
      FREE(howmany_dims);

      /* */
      return(0);
     }
     else
     {
      if (isn==1) /* backward */
      {
       if (Sci_Plan.norm) /* normalized ffti */
       {
        /* */
        double ak;

        ak=1/((double)(n[0]));

        /* must find a BLAS/calelm function to do that */
        for(i=0;i<k;i++)
        {
         ri[i]=ri[i]*ak;
         ii[i]=ii[i]*ak;
        }
       }
      }

      /* */
      for(i=0;i<nspn[0];i++)
        fftw_execute_split_dft(p,&ri[i],&ii[i],&ro[i],&io[i]);

      /* */
      FREE(dims);
      FREE(howmany_dims);
      FREE(n);FREE(nspn);

      /* in place fft*/
      if (Sci_Plan.in) LhsVar(1) = 1;
      /* out place fft*/
      else LhsVar(1) = 2;

      /* */
      return(0);
     }
    }

    /* size mxn Rhs(3)/Rhs(4)  */
    else
    {
      /* set parameters of fftw_plan_guru_split_dft */
      rank=mn3;
      if ((dims=(fftw_iodim *)MALLOC(sizeof(fftw_iodim)*rank))==NULL)
      {
       /* */
       Scierror(999,"%s: "
                    "Allocation memory error\n",
                    fname);
       /* */
       return(0);
      }

      /* */
      for(j=0;j<rank;j++)
      {
       dims[j].n = n[j];
       dims[j].is = nspn[j];
       dims[j].os = nspn[j];
      }

      howmany_rank=0;
      howmany_dims=NULL;
    }

   }
  }

  /*****************
   * Get fftw plan *
   *****************/

  /* call GetFFTWPlan */
  if ((p = GetFFTWPlan(rank, dims,
                       howmany_rank, howmany_dims,
                       ri, ii, ro, io,
                       Sci_Plan.flags,isn)) == NULL)
  {
   /* */
   Scierror(999,"%s: "
                "Allocation memory error\n",
                fname);
   /* */
   FREE(dims);
   if (howmany_dims!=NULL) FREE(howmany_dims);

   /* */
   return(0);
  }

  /* */
  FREE(dims);
  if (howmany_dims!=NULL) FREE(howmany_dims);

  /***************
   * Compute fft *
   ***************/

  if (isn==1) /* backward */
  {
   if (Sci_Plan.norm) /* normalized ffti */
   {
    /* */
    double ak;

    if (Rhs==4)
    {
     ak=1;
     for(i=0;i<mn3;i++) ak=ak*1/((double)(n[i]));
     FREE(n);FREE(nspn);
    }
    else
    {
     ak=1/((double)(m1*n1));
    }

    /* must find a BLAS/calelm function to do that */
    for(i=0;i<m1*n1;i++)
    {
     ri[i]=ri[i]*ak;
     ii[i]=ii[i]*ak;
    }
   }
  }

  /* execute FFTW plan */
  fftw_execute_split_dft(p,ri,ii,ro,io);

  /***********************************
   * Return results in lhs parameter *
   ***********************************/

  /* in place fft*/
  if (Sci_Plan.in) LhsVar(1) = 1;
  /* out place fft*/
  else LhsVar(1) = 2;

  /* */
  return(0);
}

/* fftw_flags function.
 *
 * Scilab Calling sequence :
 *   -->[a,b]=fftw_flags();
 * or
 *   -->[a,b]=fftw_flags(S);
 *
 *  a is an int scalar. (the int value of
 *                            the flag parameter of fftw)
 *  b is a string matrix.(the string values of
 *                             the flag parameter of fftw)
 *  S is a string matrix or an int or a double scalar
 *  given the value(s) of the fftw flag parameter.
 *
 *  This function gives and set the flag parameter of fftw
 *  when creating a new plan.
 *  This should be done before calling fftw function.
 *  (default is FFTW_ESTIMATE)
 */
int intsetfftwflags_c(char *fname)
{
  /* declaration of variables to store scilab parameters address */
  static int l1, m1, n1;
  SciIntMat M1;
  char **Str1;

  static int l2, m2, n2;

  static int l3, m3, n3;
  char **Str3=NULL;
  int len;

  /* please update me ! */
  static int nb_flag=22;
  static char *Str[]= {/* documented flags */
                       "FFTW_MEASURE",
                       "FFTW_DESTROY_INPUT",
                       "FFTW_UNALIGNED",
                       "FFTW_CONSERVE_MEMORY",
                       "FFTW_EXHAUSTIVE",
                       "FFTW_PRESERVE_INPUT",
                       "FFTW_PATIENT",
                       "FFTW_ESTIMATE",

                       /* undocumented beyond-guru flags */
                       "FFTW_ESTIMATE_PATIENT",
                       "FFTW_BELIEVE_PCOST",
                       "FFTW_NO_DFT_R2HC",
                       "FFTW_NO_NONTHREADED",
                       "FFTW_NO_BUFFERING",
                       "FFTW_NO_INDIRECT_OP",
                       "FFTW_ALLOW_LARGE_GENERIC",
                       "FFTW_NO_RANK_SPLITS",
                       "FFTW_NO_VRANK_SPLITS",
                       "FFTW_NO_VRECURSE",
                       "FFTW_NO_SIMD",
                       "FFTW_NO_SLOW",
                       "FFTW_NO_FIXED_RADIX_LARGE_N",
                       "FFTW_ALLOW_PRUNING"};

  static unsigned flagt[]= {/* documented flags */
                            FFTW_MEASURE,
                            FFTW_DESTROY_INPUT,
                            FFTW_UNALIGNED,
                            FFTW_CONSERVE_MEMORY,
                            FFTW_EXHAUSTIVE,
                            FFTW_PRESERVE_INPUT,
                            FFTW_PATIENT,
                            FFTW_ESTIMATE,

                            /* undocumented beyond-guru flags */
                            FFTW_ESTIMATE_PATIENT,
                            FFTW_BELIEVE_PCOST,
                            FFTW_NO_DFT_R2HC,
                            FFTW_NO_NONTHREADED,
                            FFTW_NO_BUFFERING,
                            FFTW_NO_INDIRECT_OP,
                            FFTW_ALLOW_LARGE_GENERIC,
                            FFTW_NO_RANK_SPLITS,
                            FFTW_NO_VRANK_SPLITS,
                            FFTW_NO_VRECURSE,
                            FFTW_NO_SIMD,
                            FFTW_NO_SLOW,
                            FFTW_NO_FIXED_RADIX_LARGE_N,
                            FFTW_ALLOW_PRUNING};

  unsigned flagv=0;

  /* local counter variable */
  int i,j;

  /* check min/max rhs parameters of scilab funcion */
  CheckRhs(0,1);

  /* */
  if (Rhs==0)
  {

  }
  else
  {
   /* int */
   if (VarType(1)==8)
   {
    /* */
    GetRhsVar(1, "I", &m1, &n1, &M1);

    /* */
    CheckDims(1,m1,n1,1,1);

    /* */
    Sci_Plan.flags=((int *)M1.D)[0];

   }
   else if (VarType(1)==1)  /* double */
   {
    /* */
    GetRhsVar(1, "d", &m1, &n1, &l1);

    /* */
    CheckDims(1,m1,n1,1,1);

    /* */
    Sci_Plan.flags=(int)*stk(l1);
   }
   else if (VarType(1)==10)  /* string */
   {
    /* */
    GetRhsVar(1,"S",&m1,&n1,&Str1);

    /* pass */
    for (j=0;j<m1*n1;j++)
    {
     for (i=0;i<nb_flag;i++)
     {
      if (strcmp(Str1[j],Str[i])==0) break;
     }
     if (i==nb_flag)
     {
      Scierror(999,"%s: "
                   "Bad flag %s for Rhs(1) parameter\n",
                  fname,Str1[j]);
      return(0);
     }
     else
     {
      if (i>0) flagv=flagv|(1U << (i-1));
     }
    }
    Sci_Plan.flags = flagv;
    FreeRhsSVar(Str1);
   }

   else
   {
    Scierror(53,"%s: "
                "Bad type for Rhs(1) parameter\n",
                fname);

    return(0);
   }
  }

  /* return value of Sci_Plan.flags in position 2 */
  m2 = 1;
  n2 = m2;
  l2 = I_INT32;
  CreateVar(2,"I",&m2,&n2,&l2);
  *istk(l2)=(int) Sci_Plan.flags;

  /* */
  j=0;

  for (i=0;i<nb_flag;i++)
  {
   /*  */
   if((Sci_Plan.flags&flagt[i])==flagt[i])
   {
    j++;
    if ((Str3=(char **)REALLOC(Str3,sizeof(char *)*j)) == NULL)
    {
      Scierror(999,"%s: "
                   "Memory allocation error\n",
                   fname);
      return(0);
    }
    len=strlen(Str[i]);
    if ((Str3[j-1]=(char *)MALLOC(sizeof(char)*(len+1))) == NULL)
    {
      Scierror(999,"%s: "
                   "Memory allocation error\n",
                   fname);
      return(0);
    }
    sprintf(Str3[j-1],"%s",Str[i]);
   }
  }

  /* */
  n1=1;
  CreateVarFromPtr( 3, "S", &j, &n1, Str3);

  /* */
  LhsVar(1)=2;
  LhsVar(2)=3;

  /* */
  for (i=0;i<j;i++) FREE(Str3[i]);
  FREE(Str3);

  /* */
  return(0);
}

/* fftw_inplace function.
 *
 * Scilab Calling sequence :
 *   -->[a]=fftw_inplace();
 * or
 *   -->[a]=fftw_inplace(b);
 *
 * with a and b boolean scalar.
 *
 * give and set if fftw works in place.
 * (default is true)
 */
int intsetfftwinplace_c(char *fname)
{
  static int l1, l2, m1, n1;

  CheckRhs(0,1);

  if (Rhs==1)
  {
   GetRhsVar(1,"b",&m1,&n1,&l1);

   CheckDims(1,m1,n1,1,1);

   Sci_Plan.in=*istk(l1);
  }

  /* */
  m1=1;n1=m1;
  CreateVar(2, "b",&m1,&n1,&l2)
  *istk(l2)=Sci_Plan.in;

  /* */
  LhsVar(1)=2;

  /* */
  return(0);
}

/* fftw_norm function.
 *
 * Scilab Calling sequence :
 *   -->[a]=fftw_norm();
 * or
 *   -->[a]=fftw_norm(b);
 *
 * with a and b boolean scalar.
 *
 * give and set if scilab fftw function
 * compute normalized ifft.
 * (default is true)
 *
 */
int intsetfftwnorm_c(char *fname)
{
  static int l1, l2, m1, n1;

  CheckRhs(0,1);

  if (Rhs==1)
  {
   GetRhsVar(1,"b",&m1,&n1,&l1);

   CheckDims(1,m1,n1,1,1);

   Sci_Plan.norm=*istk(l1);
  }

  /* */
  m1=1;n1=m1;
  CreateVar(2, "b",&m1,&n1,&l2)
  *istk(l2)=Sci_Plan.norm;

  /* */
  LhsVar(1)=2;

  /* */
  return(0);
}

/* pushfftw_plan function.
 *
 * Scilab Calling sequence :
 *   -->[a]=pushfftw_plan();
 *
 *  a is a double scalar given the
 *  current plan number stored in
 *  Sci_Plan structure.
 *
 *  store the last created plan in memory
 *  to reuse-it.
 *  It should be used when fftw is called
 *  many times with many set of parameters...
 *
 */
int intpushfftwplan_c(char *fname)
{
  /* */
  static int l1, m1, n1;

  /* declaration of local variable */
  fftw_plan p;
  int curp;

  CheckLhs(0,1);

  if (Sci_Plan.p!=NULL)
  {
   /* */
   if(PushFTTWPlan())
   {
    /* get current plan number*/
    curp = Sci_Plan.nb_CPlan;

    /* create a new plan */
    p = fftw_plan_guru_split_dft(Sci_Plan.rank[curp],
                                 Sci_Plan.dims[curp],
                                 Sci_Plan.howmany_rank[curp],
                                 Sci_Plan.howmany_dims[curp],
                                 Sci_Plan.io[0],
                                 Sci_Plan.io[1],
                                 Sci_Plan.io[2],
                                 Sci_Plan.io[3],
                                 Sci_Plan.sflags[curp]);

    /* store new plan and its parameters in Sci_Plan */
    if ((UpdateFTTWPlan(p,
                        Sci_Plan.rank[curp],
                        Sci_Plan.dims[curp],
                        Sci_Plan.howmany_rank[curp],
                        Sci_Plan.howmany_dims[curp],
                        Sci_Plan.isn[curp]))==0)
    {
     /* */
     fftw_destroy_plan(p);
     /* */
     p=NULL;
    }
   }
  }

  /* */
  m1=1;n1=m1;
  CreateVar( 1, "d", &m1, &n1, &l1);
  *stk(l1)=Sci_Plan.nb_CPlan;

  /* */
  LhsVar(1)=1;

  /* */
  return(0);
}

/* freefftw_plan function.
 *
 * Scilab Calling sequence :
 *   -->[a]=freefftw_plan();
 *
 *  a is an double scalar given the
 *  current plan number stored in
 *  Sci_Plan structure. (always 0!)
 *
 * free all the static structure Sci_Plan.
 *
 */
int intfreefftwplan_c(char *fname)
{
  static int l1, m1, n1;

  CheckLhs(0,1);

  /* */
  FreeFTTWPlan();

  /* */
  m1=1;n1=m1;
  CreateVar( 1, "d", &m1, &n1, &l1);
  *stk(l1)=Sci_Plan.nb_CPlan;

  /* */
  LhsVar(1)=1;

  /* */
  return(0);
}

/**********
 * gateway
 **********/

static GenericTable Tab[]=
{
 {(Myinterfun)sci_gateway,intfftw_c,
   "fftw"},

 {(Myinterfun)sci_gateway,intsetfftwflags_c,
   "fftw_flags"},

 {(Myinterfun)sci_gateway,intsetfftwinplace_c,
   "fftw_inplace"},

 {(Myinterfun)sci_gateway,intsetfftwnorm_c,
   "fftw_norm"},

 {(Myinterfun)sci_gateway,intpushfftwplan_c,
   "pushfftw_plan"},

 {(Myinterfun)sci_gateway,intfreefftwplan_c,
   "freefftw_plan"},
};

int C2F(intfftw3)()
{
 Rhs = Max(0, Rhs);
 (*(Tab[Fin-1].f))(Tab[Fin-1].name,Tab[Fin-1].F);
  return 0;
}

